import numpy as np
import torch
import pandas as pd
from itertools import product
from scipy import stats
from torch import nn


def get_blosum62_index(char):
    blosum62 = {
    'A': [ 4, -1, -2, -2, 0, -1, -1, 0, -2, -1, -1, -1, -1, -3, -1, 1, 0, -3, -2, 0],
    'R': [-1, 5, 0, -2, -3, 1, 0, -2, 0, -3, -2, 2, -1, -3, -2, -1, -1, -3, -2, -3],
    'N': [-2, 0, 6, 1, -3, 0, 0, 0, 1, -3, -3, 0, -2, -3, -2, 1, 0, -4, -2, -3],
    'D': [-2, -2, 1, 6, -3, 0, 2, -1, -1, -3, -4, -1, -3, -4, -1, 0, -1, -4, -3, -3],
    'C': [ 0, -3, -3, -3, 9, -3, -4, -3, -3, -1, -1, -3, -1, -2, -3, -1, -1, -2, -2, -1],
    'Q': [-1, 1, 0, 0, -3, 5, 2, -2, 0, -3, -2, 1, 0, -3, -1, 0, -1, -2, -1, -2],
    'E': [-1, 0, 0, 2, -4, 2, 5, -2, 0, -3, -3, 1, -2, -3, -1, 0, -1, -3, -2, -2],
    'G': [ 0, -2, 0, -1, -3, -2, -2, 6, -2, -4, -4, -2, -3, -4, -2, 0, -2, -4, -3, -2],
    'H': [-2, 0, 1, -1, -3, 0, 0, -2, 8, -3, -3, -1, -2, -1, -2, -1, -2, -2, 2, -3],
    'I': [-1, -3, -3, -3, -1, -3, -3, -4, -3, 4, 2, -3, 1, 0, -3, -2, -1, -3, -1, 3],
    'L': [-1, -2, -3, -4, -1, -2, -3, -4, -3, 2, 4, -2, 2, 0, -3, -2, -1, -2, -1, 1],
    'K': [-1, 2, 0, -1, -3, 1, 1, -2, -1, -3, -2, 5, -1, -3, -1, 0, -1, -3, -2, -2],
    'M': [-1, -1, -2, -3, -1, 0, -2, -3, -2, 1, 2, -1, 5, 0, -2, -1, -1, -1, -1, 1],
    'F': [-3, -3, -3, -4, -2, -3, -3, -4, -1, 0, 0, -3, 0, 6, -4, -2, -2, 1, 3, -1],
    'P': [-1, -2, -2, -1, -3, -1, -1, -2, -2, -3, -3, -1, -2, -4, 7, -1, -1, -4, -3, -2],
    'S': [ 1, -1, 1, 0, -1, 0, 0, 0, -1, -2, -2, 0, -1, -2, -1, 4, 1, -3, -2, -2],
    'T': [ 0, -1, 0, -1, -1, -1, -1, -2, -2, -1, -1, -1, -1, -2, -1, 1, 5, -2, -2, 0],
    'W': [-3, -3, -4, -4, -2, -2, -3, -4, -2, -3, -2, -3, -1, 1, -4, -3, -2, 11, 2, -3],
    'Y': [-2, -2, -2, -3, -2, -1, -2, -3, 2, -1, -1, -2, -1, 3, -3, -2, -2, 2, 7, -1],
    'V': [ 0, -3, -3, -3, -1, -2, -2, -2, -3, 3, 1, -2, 1, -1, -2, -2, 0, -3, -1, 4]
    }
    keys_list = list(blosum62.keys())
    return keys_list.index(char)


# 对氨基酸序列进行编码blosum62
def encoding(aa):
    blosum = np.array([
        [4, -1, -2, -2, 0, -1, -1, 0, -2, -1, -1, -1, -1, -2, -1, 1, 0, -3, -2, 0],
        [-1, 5, 0, -2, -3, 1, 0, -2, 0, -3, -2, 2, -1, -3, -2, -1, -1, -3, -2, -3],
        [-2, 0, 6, 1, -3, 0, 0, 0, 1, -3, -3, 0, -2, -3, -2, 1, 0, -4, -2, -3],
        [-2, -2, 1, 6, -3, 0, 2, -1, -1, -3, -4, -1, -3, -3, -1, 0, -1, -4, -3, -3],
        [0, -3, -3, -3, 9, -3, -4, -3, -3, -1, -1, -3, -1, -2, -3, -1, -1, -2, -2, -1],
        [-1, 1, 0, 0, -3, 5, 2, -2, 0, -3, -2, 1, 0, -3, -1, 0, -1, -2, -1, -2],
        [-1, 0, 0, 2, -4, 2, 5, -2, 0, -3, -3, 1, -2, -3, -1, 0, -1, -3, -2, -2],
        [0, -2, 0, -1, -3, -2, -2, 6, -2, -4, -4, -2, -3, -3, -2, 0, -2, -2, -3, -3],
        [-2, 0, 1, -1, -3, 0, 0, -2, 8, -3, -3, -1, -2, -1, -2, -1, -2, -2, 2, -3],
        [-1, -3, -3, -3, -1, -3, -3, -4, -3, 4, 2, -3, 1, 0, -3, -2, -1, -3, -1, 3],
        [-1, -2, -3, -4, -1, -2, -3, -4, -3, 2, 4, -2, 2, 0, -3, -2, -1, -2, -1, 1],
        [-1, 2, 0, -1, -3, 1, 1, -2, -1, -3, -2, 5, -1, -3, -1, 0, -1, -3, -2, -2],
        [-1, -1, -2, -3, -1, 0, -2, -3, -2, 1, 2, -1, 5, 0, -2, -1, -1, -1, -1, 1],
        [-2, -3, -3, -3, -2, -3, -3, -3, -1, 0, 0, -3, 0, 6, -4, -2, -2, 1, 3, -1],
        [-1, -2, -2, -1, -3, -1, -1, -2, -2, -3, -3, -1, -2, -4, 7, -1, -1, -4, -3, -2],
        [1, -1, 1, 0, -1, 0, 0, 0, -1, -2, -2, 0, -1, -2, -1, 4, 1, -3, -2, -2],
        [0, -1, 0, -1, -1, -1, -1, -2, -2, -1, -1, -1, -1, -2, -1, 1, 5, -2, -2, 0],
        [-3, -3, -4, -4, -2, -2, -3, -2, -2, -3, -2, -3, -1, 1, -4, -3, -2, 11, 2, -3],
        [-2, -2, -2, -3, -2, -1, -2, -3, 2, -1, -1, -2, -1, 3, -3, -2, -2, 2, 7, -1],
        [0, -3, -3, -3, -1, -2, -2, -3, -3, 3, 1, -2, 1, -1, -2, -2, 0, -3, -1, 4]
    ])
    new_seq = []
    for i in aa:
        new_seq.append(blosum[get_blosum62_index(i)])
    flat_list = [item for sublist in new_seq for item in sublist]
    return flat_list


# 构建模型和损失函数进行训练，返回训练后的模型/或者sample一些target——train，得到提取后的特征，获得单类特征向量
# 模型选择，初步使用了transformer的编码器
# 这里之后要预训练一个transformer，直接调用预训练的参数来初始化
class TransEncoder(nn.Module):
    def __init__(self):
        super(TransEncoder, self).__init__()
        # embedding已经在外部写好了
        # nhead should be divided by d_model
        encoder_layer = nn.TransformerEncoderLayer(d_model=20, nhead=5)
        self.TransformerEncoder = nn.TransformerEncoder(encoder_layer, num_layers=6)

    def forward(self, x):
        # x的shape(batch-size, seq-len95, aa-dim20)
        out = self.TransformerEncoder(x)
        # 将三维转变为二维(batch-size, seq-len*aa-dim1900)
        out = out.view(out.size(0), -1)
        return out


# 定义ref model,两个model共享encoder
class RefEncoder(nn.Module):
    # num_classes需要读取ref的种类
    def __init__(self, encoder, num_classes, input_dim=1600):
        super(RefEncoder, self).__init__()
        self.encoder = encoder
        # 实际fc层更新？
        self.fc = nn.Linear(input_dim, num_classes)

    def forward(self, x):
        # x的shape(batch-size, seq-len95, aa-dim20)
        out = self.encoder(x)
        # 将三维转变为二维(batch-size, seq-len*aa-dim1900)
        # out = out.view(out.size(0), -1)
        # 过分类头
        out = self.fc(out)
        return out


# 定义target model,两个model共享encoder
class TarEncoder(nn.Module):
    # num_classes需要读取ref的种类
    def __init__(self, encoder):
        super(TarEncoder, self).__init__()
        self.encoder = encoder

    def forward(self, x):
        # x的shape(batch-size, seq-len95, aa-dim20)
        out = self.encoder(x)
        # 将三维转变为二维(batch-size, seq-len*aa-dim1900)
        # out = out.view(out.size(0), -1)
        return out


# input_seq是输入序列，df是对应物种的所有数据
def seq_fre_get(input_seq, data_path):
    # 里面存储96个list/set，代表每一个位置可能的位点
    seq_list = []
    # 先将原数据保留到每一个位置的可能位点中
    for i in range(len(input_seq)):
        each_aa = [input_seq[i]]
        seq_list.append(each_aa)
    # print(seq_list)
    # 统计对应物种
    df = pd.read_csv(data_path, low_memory=False)
    df = pd.DataFrame(df)
    # 对于首列数据单独统计
    seq_first = df[df['X_1'].map(lambda x:x != '#')]['X_1']
    seq_first_fre = pd.value_counts(seq_first, normalize=True)
    seq_first_fre = seq_first_fre.to_dict()
    # 统计此列数据并确定可能位点氨基酸
    counter1 = 0
    sum_fre = 0
    # 存储统计要生成物种所保留的氨基酸
    ans_tar = []
    for key, value in seq_first_fre.items():
        sum_fre = sum_fre + value
        counter1 = counter1 + 1
        # 该位点要包括输入和累积和在0.9的氨基酸
        ans_tar = ans_tar + [key]
        # seq_list[0] = seq_list[0] + [key]
        # 当达到一定频率就不再统计
        if sum_fre >= 0.85:
            # 如果该物种此位置一种元素就占极大比例，则此位点固定为这一元素，不考虑输入的情况
            if counter1 == 1:
                seq_list[0] = ans_tar
            # 如果鼠在人的结果里面，那就只保留鼠
            elif input_seq[0] in ans_tar:
                seq_list[0] = [input_seq[0]]
            else:
                # 第一位点要删除因为输入导致的#
                # print(seq_list[0])
                if input_seq[0] == '#':
                    if len(ans_tar) >= 2:
                        ans_tar = ans_tar[:2]
                    seq_list[0] = ans_tar
                else:
                    if len(ans_tar) >= 2:
                        ans_tar = ans_tar[:2]
                    seq_list[0] = ans_tar
            break

    # print(seq_list[0])
    # 删掉第一列和物种列，遍历所有位点
    df = df.drop(labels=['X_1', 'specie'], axis=1)
    # 记录列数
    row_counter = 1
    for index, row in df.iteritems():
        seq_fre = pd.value_counts(row, normalize=True)
        # print(seq_fre)
        # seq_fre中为人的每一个位点的氨基酸频率统计情况
        seq_fre = seq_fre.to_dict()
        # 统计此列数据并确定可能位点氨基酸
        counter1 = 0
        sum_fre = 0
        ans_tar = []
        for key, value in seq_fre.items():
            sum_fre = sum_fre + value
            counter1 = counter1 + 1
            # 该位点要包括输入和累积和在0.9的氨基酸
            ans_tar = ans_tar + [key]
            # seq_list[row_counter] = seq_list[row_counter] + [key]
            # 当达到一定频率就不再统计
            if sum_fre >= 0.9:
                # 如果该物种此位置一种元素就占极大比例，则此位点固定为这一元素，不考虑输入的情况
                if counter1 == 1:
                    seq_list[row_counter] = [key]
                # 如果鼠在人的结果里面，那就只保留鼠+人的频率最高，如果人就是鼠的频率最高，后面会去重
                elif input_seq[row_counter] in ans_tar:
                    seq_list[row_counter] = [input_seq[row_counter]]+[ans_tar[0]]
                else:
                    # 某些位置氨基酸比较分散，只保留频率最高的两位+鼠（后面会去重）
                    if len(ans_tar) >= 2:
                        ans_tar = ans_tar[:2]
                    seq_list[row_counter] = ans_tar
                break
        row_counter = row_counter + 1
    # 去重
    for i in range(len(seq_list)):
        seq_list[i] = list(set(seq_list[i]))
    ans_num = 1
    print('Every bit of possible amino acids:')
    for i in range(len(seq_list)):
        print(seq_list[i])
        ans_num = ans_num * len(seq_list[i])
    return seq_list


def gen_seq(seq_list):
    gen_list = []
    # df = pd.DataFrame()
    j = 1
    for i in product(*seq_list):
        j = j + 1
        gen_aa = ''.join(i)
        gen_list.append(gen_aa)
        if j == 1000:
            break
    return gen_list


def seq_score(input_cdr, gen_list, model, tar_emb, out_path):
    scores = []
    model.eval()
    with torch.no_grad():
        for seq in gen_list:
            seq = torch.tensor(encoding(seq)).float()
            seq = torch.reshape(seq, (-1, 80, 20))
            output = model(seq)
            cor = stats.pearsonr(tar_emb, output[0].cpu().numpy())[0]
            scores.append(cor)
    seq_com = []
    for seq in gen_list:
        aa = seq[:23] + input_cdr[0] + seq[23:38] + input_cdr[1] + seq[38:70] + input_cdr[2] + seq[70:]
        seq_com.append(aa)
    df = pd.DataFrame()
    df['seq'] = seq_com[0:-1]
    df['score'] = scores[0:-1]
    df.sort_values(by='score', inplace=True, ascending=False)
    seq_highest = df['seq'].tolist()[0]
    with open('structure_seq.txt', 'w') as f:
        f.write(seq_highest + '\n')
    input_seq = seq_com[-1]
    scores = scores[-1]
    df.loc[len(df.index)] = [input_seq, scores]
    df.to_csv(out_path, sep=',', index=False, header=True)


def read_input(specie_path, input_path):
    with open(specie_path, "r") as f:
        tar_specie = f.read()
    ans = []
    with open(input_path, "r") as f:
        for line in f.readlines():
            line = line.strip('\n')  # 去除文本中的换行符
            ans.append(line)
    input_seq_fr = ans[0]
    input_cdr = ans[1:]
    return tar_specie, input_seq_fr, input_cdr


if __name__ == '__main__':
    specie_path = './specie_target.txt'
    input_path = "./input_seq.txt"
    tar_specie, input_seq_fr, input_cdr = read_input(specie_path, input_path)
    print(tar_specie)
    print(input_seq_fr)
    print(input_cdr)
    # 以鼠源抗体人源化为例，data_path中为人的抗体csv
    data_path = './results/data/' + tar_specie + '.csv'
    # 生成FR区共96个位点每个位点可能的氨基酸
    seq_list = seq_fre_get(input_seq_fr, data_path)
    # 根据每个位点可能氨基酸组合生成所有结果
    gen_list = gen_seq(seq_list)
    gen_list.append(input_seq_fr)
    # 分别加载人的模型和特征向量
    tar_emb = np.load('./results/features/'+tar_specie+'.features.npy')
    model = torch.load('./results/models/'+tar_specie+'.model.pt', map_location='cpu')
    # 进行评分并返回生成的抗体数据（完整，包括FR和CDR区）
    seq_score(input_cdr, gen_list, model, tar_emb, 'ans_com.csv')
    print("The results of the build can be found in the ans_com file.")
    print("The last sequence is the input sequence.")
    print("The highest-rated sequence is stored in structure_seq file.")